/******************************************************************************
*                            Steffen Hofmann                                  *
*                            mcu@shotech.de                                   *
*******************************************************************************
*  $Source:  $
*-----------------------------------------------------------------------------*
*  Datum der 2. Version  :  12.03.2017 20:24:04
*  Geraete Bezeichnung   :  RFM69CW-RF-Modul
*-----------------------------------------------------------------------------*
*- $Id:  $
*-----------------------------------------------------------------------------*
*
*	Headerfile
*	----------
*	Clock Projekt RFM69CW-RF-Modul
* 
*
******************************************************************************/

#include "hw.h"
#include "clock.h"


/* Private macro -------------------------------------------------------------*/
/* Private variables ---------------------------------------------------------*/
static __IO uint32_t Tick;
volatile uint16_t error = 0;  //initialized at 0 and modified by the functions 

volatile uint16_t wTickApplikation = 1;
volatile bool bIsTimeout = false;
volatile uint32_t wRfTickCount;

u16 getError( void )
{
	return error;
}

void setTickCount( u32 dTick )
{
	wTickApplikation = 0;
	bIsTimeout = false;
	wRfTickCount = dTick;
	SYSTICK_ENABLE();
}
u32 getTickCount( void )
{
	return Tick;
}

bool getTimeout(void)
{
	return bIsTimeout;
}

/**
  * Brief   This function configures the system clock @16MHz and voltage scale 1
  *         assuming the registers have their reset value before the call.
  *         POWER SCALE   = RANGE 1
  *         SYSTEM CLOCK  = PLL MUL8 DIV2
  *         PLL SOURCE    = HSI/4
  *         FLASH LATENCY = 0
  * Param   None
  * Retval  None
  */
void SystemClock_Config(void)
{
  uint32_t tickstart;
  /* (1) Enable power interface clock */
  /* (2) Select voltage scale 1 (1.65V - 1.95V) 
         i.e. (01)  for VOS bits in PWR_CR */
  /* (3) Enable HSI in RCC-> CR */
  /* (4) Wait for HSI ready flag  */
  /* (5) Set PLL on HSI, multiply by 8 and divided by 2 */
  /* (6) Enable the PLL in RCC_CR register */
  /* (7) Wait for PLL ready flag */
  /* (8) Select PLL as system clock */
  /* (9) Wait for clock switched on PLL */
  RCC->APB1ENR |= (RCC_APB1ENR_PWREN); /* (1) */
  PWR->CR = (PWR->CR & ~(PWR_CR_VOS)) | PWR_CR_VOS_0; /* (2) */
  
  RCC->CR |= RCC_CR_HSION; /* (3) */
  tickstart = Tick;
  while ((RCC->CR & (RCC_CR_HSIRDY)) != (RCC_CR_HSIRDY)) /* (4) */
  {
    if ((Tick - tickstart ) > HSI_TIMEOUT_VALUE)
    {
      error = ERROR_HSI_TIMEOUT; /* Report an error */
      return;
    }      
  }
  
  RCC->CFGR |= RCC_CFGR_SW_HSI; /* (8) */
  tickstart = Tick;
  while ((RCC->CFGR & RCC_CFGR_SWS_HSI)  == 0) /* (9) */
  {
    if ((Tick - tickstart ) > CLOCKSWITCH_TIMEOUT_VALUE)
    {
      error = ERROR_CLKSWITCH_TIMEOUT; /* Report an error */
      return;
    }      
  }  
  
}

/**
  * Brief   This function handles SysTick Handler.
  *         It toggles the green led if the action has been performed correctly
  *         and toggles the red led coding the error number
  * Param   None
  * Retval  None
  */
void SysTick_Handler(void)
{
  static uint32_t long_counter = LONG_DELAY;
  static uint32_t short_counter = SHORT_DELAY;  
  static uint16_t error_temp = 0;
  
  
  if( wTickApplikation != 0 )
  	{
			  Tick++;
			  if (long_counter-- == 0) 
			  {
			    if(error == 0)
			    {
			      /* the following instruction can only be used if no ISR modifies GPIOC ODR
			         either by writing directly it or by using GPIOC BSRR or BRR 
			         else a toggle mechanism must be implemented using GPIOC BSRR and/or BRR
			      */
			      long_counter = LONG_DELAY;
			    }
			    else if (error != 0xFF)
			    {
			      /* red led blinks according to the code error value */
			      error_temp = (error << 1) - 1;
			      short_counter = SHORT_DELAY;
			      long_counter = LONG_DELAY << 1;
			      //GPIOA->BSRR = (1 << 5); //set red led on PA5
			      //GPIOB->BRR = (1 << 4); //switch off green led on PB4
			      
			      LED_RED_HIGH();
			      LED_GREEN_LOW();
			    }
			  }
			  if (error_temp > 0)
			  {
			    if (short_counter-- == 0) 
			    {
			      //GPIOA->ODR ^= (1 << 5); //toggle red led
			      LED_RED_TOGGLE();
			      short_counter = SHORT_DELAY;
			      error_temp--;
			    }  
			  }
		}
		else
			{
				  if( wRfTickCount != 0 )
				  	{
				      wRfTickCount--;
				    }
				    else
				  	  {
				  		   SYSTICK_DISABLE();
				  		   bIsTimeout = true;
				  	  }
				
			}
}